// Yes, this is an included .c file.  So shoot me.

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <unistd.h>
#include <sysexits.h>
#include <netdb.h>

#ifndef SPOOLDIR
// be sure that this directory exists and is writable by the
// process that invokes this code; also that it contains a "tmp"
// subdirectory (assuming tmpnam(NULL) returns "/tmp/something")
#define SPOOLDIR "/var/spool/backupmx"
// Delivery info stored in .job file.  Same filename without an
// extension contains the mail itself.
#endif

#ifndef EXT
#define EXT ".tmp"
// NOTE!  Hard coded that this is 3 chars and is replaced by ".job"
#endif

#ifndef JOBEXT
#define JOBEXT ".job"
#endif

#ifndef FIND_COMMAND
// %s is the SPOOLDIR below
#define FIND_COMMAND "find %s/tmp -name '*.job' -print"
#endif

#define MAX_STRING 1024

#ifndef FALSE
#define FALSE (0!=0)
#define TRUE (0==0)
#endif

static time_t starttime;

FILE *logfile = NULL;

/* printf-like procedure to write to fd's rather than files */
void put (int fd, char *s, ...)
{
   int string_length, rc;
   static char *buff = NULL;
   va_list ap;

   va_start (ap, s);
   {
   static FILE *nullfile = NULL;

      if (nullfile == NULL)
	 nullfile = fopen ("/dev/null", "w");	// Don't open on every call!
      if (nullfile == NULL) {
	 fprintf(stderr, "Cound not open /dev/null???!\n");
	 usleep(2500); exit (2);		// [TO DO]
      }
      string_length = vfprintf (nullfile, s, ap);
      // fclose (nullfile); // an 'atexit' handler could do this
      buff = malloc (string_length + 1);
      if (buff != NULL)
	 vsprintf (buff, s, ap);
   }
   va_end (ap);

   // *** Instead of a block write, we should write a character at a time,
   // and extract all "\r"s then insert \r in front of any \n
   // (hoping that it doesn't screw up 'content-length' in mail)
   // rc = write(fd, buff, string_length);
   {
   static char cr[2] = { '\r', '\0' };
   int i, c;

      for (i = 0; i < string_length; i++) {
	 c = buff[i];
	 if (c != '\r') {
	    if (c == '\n') {
	       write (fd, cr, 1);
               if (logfile != NULL) {write (fileno(logfile), cr, 1); fflush(logfile);}
	    }
	    write (fd, buff + i, 1);
            if (logfile != NULL) {write (fileno(logfile), buff + i, 1); fflush(logfile);}
	 }
      }
   }

   // might be better to keep track of max length and realloc if too small
   if (buff != NULL)
      free (buff);
   buff = NULL;
}

// Using a macro here allows us to replace this easily with a diagnostic version
#define debug_exit(rc) my_exit(rc, __LINE__)

int my_exit (int rc, int line)
{
   usleep (2500);
   // This is paranoia.  Keep the connection open a few seconds
   // longer in case the problems we are having with duplicate mails are due
   // to the final packet in the buffer (the return code) not being delivered
   // (ie a race condition) [It wasn't - it was a firewall bug mentioned
   // elsewhere; however this doesn't hurt!)
   exit (rc);
}

static jmp_buf RDTimeout;

int DeadMansTimeout = 60 * 6;	/* Defensive programming */

int ReadTimeout = 60 * 5;	/* Read timeout in SMTP phase - primarily
				   'waiting for command or reply' phases */

// If you want to be RFC compatible, set both of these to 60*5 ...
int CmdTimeout = 30;		/* This turns out to be essential if you want 
				   to avoid exceeding the max no of daemon
				   children active at once */

int ShortTimeout = 45;		/* Much shorter timeout used eg in the middle 
				   of reading a line. */

static void dtimer (int sig)
{
   longjmp (RDTimeout, 1);
}


int get (int fd, int timeout)
{
   char cbuff;
   int c;
   fd_set rfds;
   struct timeval tv;
   int retval;
   static int timedoutalready = FALSE;

   FD_ZERO (&rfds);
   FD_SET (fd, &rfds);

   tv.tv_sec = timeout;
   tv.tv_usec = 0;

   retval = select (fd + 1, &rfds, NULL, NULL, &tv);
   /* Don't rely on the value of tv now! */

   c = EOF;
   if ((retval != -1) && (retval != 0)) {	// retval > 0 ???
      // Unbuffered, one char at a time
      // but buffered would be OK too.
      if (read (fd, &cbuff, 1) > 0)
	 c = (cbuff & 255);	// BUG FIX 20040723
      /* FD_ISSET(0, &rfds) will be true. */
      if (logfile != NULL) {fputc(c, logfile); fflush(logfile);}
   }

   if (c == EOF) {
      if (timedoutalready == TRUE) {
	 // This is a belt-and-suspenders safety trick.  If I accidentally
	 // fail to exit after returning EOF, the next read will force an
	 // exit.
	 // (i.e. pascal-style!)
	 debug_exit (0);
      }
      timedoutalready = TRUE;
   }
   return c;
}

static jmp_buf RDTimeout;
static void dead_mans_timer (int sig)
{
   longjmp (RDTimeout, 1);
}
